from pathlib import Path
from shutil import copyfile

from invoke import Context


class LocalContext(Context):
    def put(self, local, remote=None):
        self._cp(local, remote, Path.cwd(), self.cwd)

    def get(self, remote, local=None):
        self._cp(remote, local, self.cwd, Path.cwd())

    def _parse(self, other, rel_to, change_parent=None):
        op = Path(other)
        if op.is_absolute():
            if change_parent and op.is_relative_to(change_parent):
                return Path(rel_to, op.relative_to(change_parent))
            else:
                return op
        return Path(rel_to, op)

    def _cp(self, source, target, source_rel, target_rel):
        true_target = self._parse(target or source, target_rel, source_rel)
        true_target.parent.mkdir(exist_ok=True, parents=True)
        copyfile(self._parse(source, source_rel), true_target)



import os
import secrets
from base64 import urlsafe_b64decode as b64d
from base64 import urlsafe_b64encode as b64e
from contextlib import contextmanager
from dataclasses import dataclass, field
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Callable, Dict, Optional, Union

from cryptography.fernet import Fernet
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.backends.openssl.backend import Backend
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.kdf.pbkdf2 import PBKDF2HMAC

CONF_PATH = Path.home() / ".zimmauth"
LOCAL_HOST_NAMES_ENV_VAR = "ZAUTH_LOCAL_HOST_NAMES"


@dataclass
class EncryptBase:
    iterations: int = 100_000
    backend: Backend = field(default_factory=default_backend)
    length: int = 32
    algo: Callable = hashes.SHA256
    salt_length: int = 16

    def encrypt(self, message: bytes, password: str) -> str:
        salt = secrets.token_bytes(self.salt_length)
        key = self._derive_key(password, salt)
        token = b64d(Fernet(key).encrypt(message))
        return b64e(b"%b%b" % (salt, token)).hex(" ")

    def decrypt(self, hex_token: str, password: str) -> bytes:
        decoded = b64d(bytes.fromhex(hex_token))
        sl = self.salt_length
        salt, token = decoded[:sl], b64e(decoded[sl:])
        key = self._derive_key(password, salt)
        return Fernet(key).decrypt(token)

    def encrypt_str(self, message: str, password: str) -> str:
        return self.encrypt(message.encode(), password)

    def _derive_key(self, password: str, salt: bytes) -> bytes:
        kdf = PBKDF2HMAC(
            algorithm=self.algo(),
            length=self.length,
            salt=salt,
            iterations=self.iterations,
            backend=self.backend,
        )
        return b64e(kdf.derive(password.encode()))


@dataclass
class S3Auth:
    key: str
    secret: str
    endpoint: Optional[str] = None

    def __repr__(self) -> str:
        return f"Auth {self.key[:2]}"


@dataclass
class S3Remote:
    bucket_name: str
    auth: S3Auth

    def to_dvc_conf(self, _):
        d = {
            "url": f"s3://{self.bucket_name}",
            "access_key_id": self.auth.key,
            "secret_access_key": self.auth.secret,
        }
        if self.auth.endpoint:
            d["endpointurl"] = self.auth.endpoint
        return d


@dataclass
class SSHRemote:
    host: str
    path: str = ""
    user: str = field(default_factory=os.getlogin)
    port: Optional[int] = None
    rsa_key: Optional[str] = None

    def __repr__(self) -> str:
        return f"SSH: {self.host}:/{self.path}"

    def to_dvc_conf(self, key_store: Path):
        if self.host in _get_local_names():
            return {"url": self.path}
        d = {"url": f"ssh://{self.host}{self.path}", "user": self.user}
        if self.port:
            d["port"] = self.port
        if self.rsa_key:
            d["keyfile"] = (key_store / self.rsa_key).as_posix()
        return d


class ZimmAuth:
    def __init__(self, dic: dict) -> None:
        def _pop(key):
            return dic.pop(key, {})

        auths = {k: S3Auth(**v) for k, v in _pop("keys").items()}
        _ssh_hosts = _pop("ssh")
        self._rsa_keys = _pop("rsa-keys")

        self._s3_remotes: Dict[str, S3Remote] = {}
        self._ssh_remotes: Dict[str, SSHRemote] = {}

        for k, r in dic.items():
            if "key" in r.keys():
                self._s3_remotes[k] = S3Remote(k, auths[r["key"]])
            if "connection" in r.keys():
                ssh_conf = _ssh_hosts[r.pop("connection")]
                remote = SSHRemote(**ssh_conf, **r)
                self._ssh_remotes[k] = remote

    def get_auth(self, remote_id: str) -> S3Auth:
        return self._s3_remotes[remote_id].auth

    @contextmanager
    def get_fabric_connection(self, remote_name: str):
        from fabric import Connection

        from .local_conn import LocalContext

        tmp_dir = TemporaryDirectory()
        ssh_rem = self._ssh_remotes[remote_name]
        if ssh_rem.host in _get_local_names():
            ctx = LocalContext()
        else:
            key_path = Path(tmp_dir.name, "key")
            c_kwargs = {}

            if ssh_rem.rsa_key:
                key = self._rsa_keys[ssh_rem.rsa_key]
                key_path.write_text(key)
                c_kwargs["key_filename"] = key_path.absolute().as_posix()

            ctx = Connection(
                host=ssh_rem.host,
                user=ssh_rem.user,
                port=ssh_rem.port,
                connect_kwargs=c_kwargs,
            )
        with ctx.cd(ssh_rem.path):
            yield ctx
        tmp_dir.cleanup()

    def dump_dvc(self, local=True, key_store=CONF_PATH):
        from dvc.config import Config

        conf = Config(Path(".dvc"))
        with conf.edit("local" if local else "global") as ced:
            ced["remote"] = {
                k: v.to_dvc_conf(key_store) for k, v in self._remotes.items()
            }
        self._dump_keys(key_store)

    def get_boto_bucket(self, remote_name):
        import boto3

        s3_rem = self._s3_remotes[remote_name]
        auth = s3_rem.auth

        return boto3.resource(
            "s3",
            endpoint_url=auth.endpoint,
            aws_access_key_id=auth.key,
            aws_secret_access_key=auth.secret,
        ).Bucket(s3_rem.bucket_name)

    @classmethod
    def from_env(cls, dic_env_var: str, pw_env_var: Optional[str] = None):
        import toml

        toml_str = os.environ.get(dic_env_var, "")
        if pw_env_var:
            assert toml_str, f"Nothing given for {dic_env_var}"
            toml_str = EncryptBase().decrypt(toml_str, os.environ[pw_env_var]).decode()

        return cls(toml.loads(toml_str))

    @staticmethod
    def dumps_dict(dic: dict, password: str):
        import toml

        return EncryptBase().encrypt_str(toml.dumps(dic), password)

    def _dump_keys(self, root: Path):
        root.mkdir(exist_ok=True, parents=True)
        for key_id, keystr in self._rsa_keys.items():
            key_path = root / key_id
            key_path.write_text(keystr)
            key_path.chmod(0o600)

    @property
    def _remotes(self) -> Dict[str, Union[S3Remote, SSHRemote]]:
        return {**self._s3_remotes, **self._ssh_remotes}


def _get_local_names():
    return os.environ.get(LOCAL_HOST_NAMES_ENV_VAR, "").split(";")



import os
import tempfile
from pathlib import Path

import boto3
import moto
from cryptography.fernet import InvalidToken
from cryptography.hazmat.backends import default_backend as crypto_default_backend
from cryptography.hazmat.primitives import serialization as crypto_serialization
from cryptography.hazmat.primitives.asymmetric import rsa
from dvc.repo import Repo
from mockssh.server import Server
from pytest import fixture, raises

from zimmauth.core import LOCAL_HOST_NAMES_ENV_VAR, EncryptBase, ZimmAuth

TEST_USER = "zuzzer"
TEST_HOST = "zhost"


@fixture(scope="session")
def private_key_path():
    key = rsa.generate_private_key(
        backend=crypto_default_backend(), public_exponent=65537, key_size=2048
    )

    private_key = key.private_bytes(
        crypto_serialization.Encoding.PEM,
        crypto_serialization.PrivateFormat.TraditionalOpenSSL,
        crypto_serialization.NoEncryption(),
    )
    with tempfile.TemporaryDirectory() as tmp_dir:
        key_path = Path(tmp_dir, "pkey")
        key_path.write_bytes(private_key)
        yield key_path


@fixture(scope="session")
def server(private_key_path: Path):
    users = {TEST_USER: private_key_path.as_posix()}
    with Server(users) as s:
        yield s


@fixture(scope="session")
def zauth_root():
    with tempfile.TemporaryDirectory() as tmp_dir:
        yield Path(tmp_dir)


@fixture(scope="session")
def far_file():
    with tempfile.TemporaryDirectory() as tmp_path:
        ff = Path(tmp_path, "farf")
        ff.write_text("Faaar")
        yield ff


@fixture
def zauth(server: Server, private_key_path: Path, zauth_root: Path, tmp_path, far_file):
    test_dic = {
        "keys": {
            "s3-key-name-1": {"key": "XYZ", "secret": "XXX"},
            "s3-key-name-2": {"key": "XYZ", "secret": "XXX", "endpoint": "http"},
        },
        "rsa-keys": {"rsa-key-name": private_key_path.read_text()},
        "ssh": {
            "ssh-name-1": {
                "host": Server.host,
                "port": server.port,
                "user": TEST_USER,
                "rsa_key": "rsa-key-name",
            },
            "ssh-name-2": {
                "host": TEST_HOST,
                "rsa_key": "rsa-key-name",
                "user": TEST_USER,
            },
        },
        "bucket-1": {"key": "s3-key-name-1"},
        "bucket-2": {"key": "s3-key-name-2"},
        "ssh-conn-1": {"connection": "ssh-name-1", "path": zauth_root.as_posix()},
        "ssh-conn-2": {"connection": "ssh-name-1"},
        "ssh-conn-3": {"connection": "ssh-name-2", "path": far_file.parent.as_posix()},
    }
    return ZimmAuth(test_dic)


@fixture
def idiotic():
    # https://github.com/aio-libs/aiobotocore/issues/755
    # https://github.com/fsspec/s3fs/issues/357

    import aiobotocore.awsrequest
    import botocore.awsrequest
    import moto.core.botocore_stubber

    _OldBotoC = botocore.awsrequest.AWSResponse.content
    _Old = moto.core.botocore_stubber.MockRawResponse

    class RHMockResp(moto.core.botocore_stubber.MockRawResponse):
        raw_headers = {b"x-amz-bucket-region": b"region"}.items()

        async def read(self):
            return super().read()

    moto.core.botocore_stubber.MockRawResponse = RHMockResp
    botocore.awsrequest.AWSResponse.content = (
        aiobotocore.awsrequest.AioAWSResponse.content
    )
    botocore.awsrequest.AWSResponse._content_prop = (
        aiobotocore.awsrequest.AioAWSResponse._content_prop
    )
    yield
    botocore.awsrequest.AWSResponse.content = _OldBotoC
    moto.core.botocore_stubber.MockRawResponse = _Old


@fixture(scope="session")
def zauth_bucket():
    with moto.mock_s3():
        conn = boto3.resource("s3")
        conn.create_bucket(
            Bucket="bucket-1",
            CreateBucketConfiguration={"LocationConstraint": "us-west-1"},
        )
        yield conn


def test_ssh(zauth: ZimmAuth, zauth_root: Path):
    with zauth.get_fabric_connection("ssh-conn-1") as conn:
        assert conn.cwd == zauth_root.as_posix()

    with zauth.get_fabric_connection("ssh-conn-2") as conn:
        assert conn.cwd == ""
    assert "XXX" not in zauth._remotes.__repr__()


def test_env_loading():
    dic = {
        "rsa-keys": {"abc": "XYZ"},
        "keys": {"s31": {"key": "xy", "secret": "123"}},
        "buc1": {"key": "s31"},
    }
    test_pw = "Pw742"
    os.environ["__ZDIC"] = ZimmAuth.dumps_dict(dic, test_pw)
    os.environ["__ZPW"] = test_pw
    zauth = ZimmAuth.from_env("__ZDIC", "__ZPW")
    assert "abc" in zauth._rsa_keys.keys()
    assert zauth.get_auth("buc1").secret == "123"


def test_dvc(zauth: ZimmAuth, tmp_path: Path, far_file: Path, zauth_bucket, idiotic):
    cwd = Path.cwd()
    os.chdir(tmp_path)
    os.environ[LOCAL_HOST_NAMES_ENV_VAR] = TEST_HOST
    try:
        Repo.init(no_scm=True)
        zauth.dump_dvc(key_store=tmp_path / "keys")
        repo = Repo()
        kp = Path("sg")
        kp.write_text("ABC")
        repo.add(kp.as_posix())
        repo.push([kp.as_posix()], remote="ssh-conn-1")
        # repo.push([kp.as_posix()], remote="bucket-1") TODO maybe this will work again
        repo.push([kp.as_posix()], remote="ssh-conn-3")
        assert ["90"] == [f.name for f in far_file.parent.iterdir() if f != far_file]

    finally:
        os.chdir(cwd)


def test_boto(zauth: ZimmAuth, zauth_bucket: boto3, tmp_path: Path):
    bucket = zauth.get_boto_bucket("bucket-1")
    fpath = tmp_path / "testfile"
    dpath = tmp_path / "downloaded"
    fpath.write_text("fing")
    bucket.upload_file(fpath.as_posix(), "test-uploaded")
    bucket.download_file("test-uploaded", dpath.as_posix())
    assert dpath.read_text() == "fing"


def test_localize(zauth: ZimmAuth, tmp_path: Path, far_file: Path):
    tfile = tmp_path / "fing"
    ffile = tmp_path / "get"
    tfile.write_text("boo")
    in_cwd_file = Path("_file").absolute()
    in_cwd_file.write_text("botd")
    os.environ[LOCAL_HOST_NAMES_ENV_VAR] = TEST_HOST
    with zauth.get_fabric_connection("ssh-conn-3") as conn:
        conn.put(in_cwd_file)
        conn.put(tfile.as_posix(), "f2")
        conn.get(far_file, ffile)
    assert ffile.exists()
    assert (far_file.parent / in_cwd_file.name).read_text() == "botd"
    in_cwd_file.unlink()


def test_secret_base():
    eb = EncryptBase()
    pw = "TestPw"
    msg = "Some Msg"
    token_hex = eb.encrypt_str(msg, pw)
    assert msg.encode() == eb.decrypt(token_hex, pw)

    with raises(InvalidToken):
        eb.decrypt(token_hex, "Wrong")
